__version__ = '0.5.3'

from . import config

# try loading config, if it doesn't work then set uconfig to None
# this is needed so that strax(en) CI  tests will work even without a config file
uconfig = config.Config()

if uconfig.is_configured:
    logger = config.setup_logger(uconfig.logging_level)

else:
    uconfig = None
    logger = config.setup_logger()

from .rundb import DB, xent_collection, xe1t_collection
from .mongo_files import MongoUploader, MongoDownloader, APIUploader, APIDownloader




import unittest

from pprint import pprint
from utilix import *


class TestQuery(unittest.TestCase):

    def test_query(self):
        db = rundb.DB()
        data = db.query(page_num=1)
        #pprint(data)
        self.assertTrue(len(data) == 1000, 'Number of records: ' + str(len(data)))
    
    def test_query_paging(self):
        db = rundb.DB()
        data = db.query(page_num=1)
        self.assertTrue(len(data) == 1000, 'Number of records: ' + str(len(data)))
        data = db.query(page_num=2)
        self.assertTrue(len(data) == 1000, 'Number of records: ' + str(len(data)))
        data = db.query(page_num=3)
        self.assertTrue(len(data) == 1000, 'Number of records: ' + str(len(data)))

    def test_query_by_source(self):
        db = rundb.DB()
        data = db.query_by_source('calibration', page_num=1)
        #pprint(data)
        self.assertTrue(len(data) > 0)
    
    def test_query_by_tag(self):
        db = rundb.DB()
        data = db.query_by_tag('test', page_num=1)
        #pprint(data)
        self.assertTrue(len(data) > 0)

if __name__ == '__main__':
    unittest.main()




import os
import configparser
import logging


def setup_logger(logging_level='WARNING'):
    logger = logging.getLogger("utilix")
    ch = logging.StreamHandler()
    ch.setLevel(logging_level)
    logger.setLevel(logging_level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


class EnvInterpolation(configparser.BasicInterpolation):
    '''Interpolation which expands environment variables in values.'''

    def before_get(self, parser, section, option, value, defaults):
        return os.path.expandvars(value)


class Config:
    # singleton
    instance = None

    def __init__(self):
        if not Config.instance:
            Config.instance = Config.__Config()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    class __Config(configparser.ConfigParser):

        def __init__(self):
            config_file_path = None

            logger = setup_logger()

            if 'XENON_CONFIG' not in os.environ:
                logger.info('$XENON_CONFIG is not defined in the environment')
            if 'HOME' not in os.environ:
                logger.info('$HOME is not defined in the environment')
                if 'USERPROFILE' in os.environ:
                    # Are you on windows?
                    home_config = os.path.join(os.environ['USERPROFILE'], '.xenon_config')
                else:
                    logger.warning('USERPROFILE is not defined in the environment')
            else:
                home_config = os.path.join(os.environ['HOME'], '.xenon_config')
            xenon_config = os.environ.get('XENON_CONFIG')

            # see if there is a XENON_CONFIG environment variable
            if xenon_config:
                config_file_path = os.environ.get('XENON_CONFIG')

            # if not, then look for hidden file in HOME
            elif os.path.exists(home_config):
                config_file_path = home_config

            else:
                logger.warning(f"Could not load a configuration file. "
                     f"You can create one at {home_config}, or set a custom path using\n\n"
                     f"export XENON_CONFIG=path/to/your/config\n")

            if config_file_path:
                logger.debug('Loading configuration from %s' % (config_file_path))
            configparser.ConfigParser.__init__(self, interpolation=EnvInterpolation())

            self.config_path = config_file_path

            try:
                self.read_file(open(config_file_path), 'r')
            except FileNotFoundError as e:
                if config_file_path is not None:
                    raise RuntimeError(f'Unable to open {config_file_path}. Please see the README for an example configuration') from e
            except TypeError:
                if config_file_path is None:
                    pass
                else:
                    raise

            self.is_configured = config_file_path is not None

        def get_list(self, category, key):
            list_string = self.get(category, key)
            return [s.strip() for s in list_string.split(',')]

        @property
        def logging_level(self):
            # look for logging level in 'basic'  field in config file. Defaults to WARNING
            level = self.get('basic', 'logging_level', fallback='WARNING').upper()
            possible_levels = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
            if level not in possible_levels:
                raise RuntimeError(f"The logging level {level} is not valid. "
                                   f"Available levels are: \n{possible_levels}.\n "
                                   f"Please modify {self.config_path}")
            return level




import utilix
import pymongo

uconfig = utilix.uconfig


def test_mongo_urls_without_secondary_preferred(collection='runs', raise_errors=True, **kwargs):
    """Test that we can do a naive initiation of the rundb"""
    # Get the url for the config
    uri = 'mongodb://{user}:{pw}@{url}'
    url = uconfig.get('RunDB', 'pymongo_url')
    user = uconfig.get('RunDB', 'pymongo_user')
    pw = uconfig.get('RunDB', 'pymongo_password')
    database = uconfig.get('RunDB', 'pymongo_database')
    uri = uri.format(user=user, pw=pw, url=url)

    # Do a naive client initialization
    c = pymongo.MongoClient(uri, serverSelectionTimeoutMS=500, **kwargs)
    DB = c[database]
    coll = DB[collection]

    # This test might fail! See the corresponding error message and:
    # https://github.com/XENONnT/utilix/pull/14. The point of this test
    # is to see if we can just plug the pymongo_url into a mongo client
    # (often not the case depending on the access to one or more of the
    # mirrors).
    # If CI is ever integrated to Utilix, we should disable the
    # raise_errors for this test.
    utilix.rundb.test_collection(coll, url, raise_errors=raise_errors)


def test_mongo_urls_secondary_preferred():
    """Do it the right way, use secondary preferred"""
    # Raise errors if this fails because then we have a big problem!
    test_mongo_urls_without_secondary_preferred(
        **dict(readPreference='secondaryPreferred'), raise_errors=True)



import subprocess
import os
import tempfile
import shlex

sbatch_template = """#!/bin/bash

#SBATCH --job-name={jobname}
#SBATCH --output={log}
#SBATCH --error={log}
#SBATCH --account=pi-lgrandi
#SBATCH --qos={qos}
#SBATCH --partition={partition}
#SBATCH --mem-per-cpu={mem_per_cpu}
#SBATCH --cpus-per-task={cpus_per_task}

{job}
"""

SINGULARITY_DIR = '/project2/lgrandi/xenonnt/singularity-images'
TMPDIR = os.path.join(os.environ.get('SCRATCH', '.'), 'tmp')


def make_executable(path):
    """Make the file at path executable, see """
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2    # copy R bits to X
    os.chmod(path, mode)


def singularity_wrap(jobstring, image, bind=('/dali', '/project2', TMPDIR)):
    """Wraps a jobscript into another executable file that can be passed to singularity exec"""
    _, exec_file = tempfile.mkstemp(suffix='.sh', dir=TMPDIR)
    make_executable(exec_file)

    bind_string = " ".join([f"--bind {b}" for b in bind])
    image = os.path.join(SINGULARITY_DIR, image)
    jobstring = 'unset X509_CERT_DIR\n' + jobstring
    new_job_string = f"""cat > {exec_file} << EOF
#!/bin/bash
{jobstring}
EOF
singularity exec {bind_string} {image} {exec_file}
rm {exec_file}
"""
    return new_job_string


def submit_job(jobstring, log='job.log', partition='xenon1t', qos='xenon1t',
               account='pi-lgrandi', jobname='somejob',
               delete_file=True, dry_run=False, mem_per_cpu=1000,
               container='xenonnt-development.simg',
               cpus_per_task=1):

    os.makedirs(TMPDIR, exist_ok=True)

    if container:
        # need to wrap job into another executable
        _, exec_file = tempfile.mkstemp(suffix='.sh')
        jobstring = singularity_wrap(jobstring, container)
        jobstring = 'module load singularity\n' + jobstring

    sbatch_script = sbatch_template.format(jobname=jobname, log=log, qos=qos, partition=partition,
                                           account=account, job=jobstring, mem_per_cpu=mem_per_cpu,
                                           cpus_per_task=cpus_per_task)

    if dry_run:
        print("=== DRY RUN ===")
        print(sbatch_script)
        return

    _, file = tempfile.mkstemp(suffix='.sbatch')
    with open(file, 'w') as f:
        f.write(sbatch_script)

    command = "sbatch %s" % file
    if not delete_file:
        print("Executing: %s" % command)
    subprocess.Popen(shlex.split(command)).communicate()

    if delete_file:
        os.remove(file)


def count_jobs(string=''):
    username = os.environ.get("USER")
    output = subprocess.check_output(shlex.split("squeue -u %s" % username))
    lines = output.decode('utf-8').split('\n')
    return len([job for job in lines if string in job])




from setuptools import setup, find_packages

# Get requirements from requirements.txt, stripping the version tags
with open('requirements.txt') as f:
    requires = [
        r.split('/')[-1] if r.startswith('git+') else r
        for r in f.read().splitlines()]

with open('README.md') as file:
    readme = file.read()

with open('HISTORY.md') as file:
    history = file.read()

setup(
    name="utilix",
    version="0.5.3",
    url='https://github.com/XENONnT/utilix',
    description="User-friendly interface to various utilities for XENON users",
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=requires,
    python_requires=">=3.6",
    long_description=readme + '\n\n' + history,
)



import unittest

from pprint import pprint
from utilix import *


class TestGet(unittest.TestCase):

    def test_doc_by_number(self):
        db = rundb.DB()
        data = db.get_doc(2000)
        #pprint(data)
        self.assertTrue('data' in data)
        self.assertTrue('start' in data)
        self.assertTrue('source' in data)
    
    def test_doc_by_number_str(self):
        db = rundb.DB()
        data = db.get_doc('2000')
        #pprint(data)
        self.assertTrue('data' in data)
        self.assertTrue('start' in data)
        self.assertTrue('source' in data)

    def test_doc_by_name(self):
        db = rundb.DB()
        data = db.get_doc('170917_1819')
        #pprint(data)
        self.assertTrue('data' in data)
        self.assertTrue('start' in data)
        self.assertTrue('source' in data)
    
    def test_data_by_number(self):
        db = rundb.DB()
        data = db.get_data(2000)
        #pprint(data)
        self.assertTrue(len(data) > 0)

    def test_data_by_name(self):
        db = rundb.DB()
        data = db.get_data('170917_1819')
        #pprint(data)
        self.assertTrue(len(data) > 0)

if __name__ == '__main__':
    unittest.main()




import re
import time
import unittest

from pprint import pprint
from utilix import *


class TestUpdate(unittest.TestCase):

    def test_update(self):

        creation_place = "test-" + str(time.time())

        db = rundb.DB()
        data = db.get_data(2000)

        db.update_data(2000,
                       {'host': 'Host',
                        'checksum': 'checksum',
                        'location': 'test_location',
                        'creation_time': '2019-04-03T03:17:03.532308',
                        'type': 'test_plugin',
                        'hash': 'c793fa6223fifadsf',
                        'status': 'transferring',
                        'lineage_hash': 'c793fa6223f488e67646cf7ed06c14653ff04186',
                        'protocol': 'FileSytemBackend',
                        'creation_place': creation_place})

        # pull back and make sure the data has been updated
        found = False
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               entry['creation_place'] == creation_place:
                found = True
        self.assertTrue(found, msg="Unable to find added entry")

        # clean up test entries
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               re.search('^test-', entry['creation_place']):
                db.delete_data(2000, entry)
    

    def test_multiple_updates(self):

        creation_place = "test-" + str(time.time())

        db = rundb.DB()
        data = db.get_data(2000)

        db.update_data(2000,
                       {'host': 'Host',
                        'checksum': 'checksum',
                        'location': 'test_location',
                        'creation_time': '2019-04-03T03:17:03.532308',
                        'type': 'test_plugin',
                        'hash': 'c793fa6223f',
                        'status': 'transferring',
                        'lineage_hash': 'c793fa6223f488e67646cf7ed06c14653ff04186',
                        'protocol': 'FileSytemBackend',
                        'creation_place': creation_place})
        
        db.update_data(2000,
                       {'host': 'Host',
                        'checksum': 'checksumi23',
                        'location': 'test_location',
                        'creation_time': '2019-04-09T03:17:03.532308',
                        'type': 'test_plugin',
                        'hash': 'c793fa6223f',
                        'status': 'transferring',
                        'lineage_hash': 'c793fa6223f488e67646cf7ed06c14653ff04186',
                        'protocol': 'FileSytemBackend',
                        'creation_place': creation_place})

        # pull back and make sure the data has been updated
        found_count = 0
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               entry['creation_place'] == creation_place:
                found_count += 1
        self.assertTrue(found_count == 1, msg="More than one entry after multiple updates: " + str(found_count))

        # clean up test entries
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               re.search('^test-', entry['creation_place']):
                db.delete_data(2000, entry)
 
 
    def test_update_meta(self):

        creation_place = "test-" + str(time.time())

        db = rundb.DB()
        data = db.get_data(2000)

        db.update_data(2000,
                       {'host': 'Host',
                        'checksum': 'checksum',
                        'location': 'test_location',
                        'creation_time': '2019-04-03T03:17:03.532308',
                        'type': 'test_plugin',
                        'lineage_hash': 'c793fa6223f488e67646cf7ed06c14653ff04186',
                        'status': 'transferring',
                        'creation_place': creation_place,
                        'meta': {'chunks': [{'chunk_i': 0,
                                             'filename': '000000',
                                             'filesize': 10486921,
                                             'first_endtime': 1505672371629233400,
                                             'first_time': 1505672371629232340,
                                             'last_endtime': 1505672379919902590,
                                             'last_time': 1505672379919901570,
                                             'n': 146594,
                                             'nbytes': 36795094},
                                            {'chunk_i': 1,
                                             'filename': '000001',
                                             'filesize': 18922980,
                                             'first_endtime': 1505672380120253340,
                                             'first_time': 1505672380120252300,
                                             'last_endtime': 1505672390381289830,
                                             'last_time': 1505672390381288790,
                                             'n': 350824,
                                             'nbytes': 88056824}],
                                 'compressor': 'zstd',
                                 'data_kind': 'records',
                                 'data_type': 'records'

                        }})

        # pull back and make sure the data has been updated
        found = False
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               entry['creation_place'] == creation_place:
                found = True
        self.assertTrue(found, msg="Unable to find added entry")

        # clean up test entries
        data = db.get_data(2000)
        for entry in data:
            if 'creation_place' in entry and \
               re.search('^test-', entry['creation_place']):
                db.delete_data(2000, entry)


if __name__ == '__main__':
    unittest.main()




import os
import requests
import re
import json
import datetime
import logging
import pymongo
from warnings import warn
import time

from . import uconfig, io
from .config import setup_logger


# Config the logger:
if uconfig is not None:
    PREFIX = uconfig.get('RunDB', 'rundb_api_url', fallback=None)
    BASE_HEADERS = {'Content-Type': "application/json", 'Cache-Control': "no-cache"}
    logger = setup_logger(uconfig.logging_level)

else:
    logger = setup_logger()


class NewTokenError(Exception):
    pass


class APIError(Exception):
    pass


def Responder(func):
    def func_wrapper(*args, **kwargs):
        st = func(*args, **kwargs)
        if st.status_code != 200:
            logger.error("\n\tAPI Call was {0}\n\tReturn code: {1}\n\tReason: {2} ".format(
                args[1],
                st.status_code,
                st.text,
            ))

            if st.status_code == 401:
                raise APIError(
                    "Error 401 is an authentication error. This is likely an issue with your token. "
                    "Can you do 'rm ~/.dbtoken' and try again? ")
        return st

    return func_wrapper


class Token:
    """
    Object handling tokens for runDB API access.

    """
    token_string = None
    user = None
    creation_time = None

    def __init__(self, path):
        self.path = path

        # if token path exists, read it in. Otherwise make a new one
        if os.path.exists(path):
            logger.debug(f'Token exists at {path}')
            with open(path) as f:
                try:
                    json_in = json.load(f)
                except json.JSONDecodeError as e:
                    raise RuntimeError(
                        f'Cannot open {path}, please report to https://github.com/XENONnT/utilix/issues. '\
                        f'To continue do "rm {path}" and restart notebook/utilix') from e
                self.token_string = json_in['string']
                self.creation_time = json_in['creation_time']
            # some old token files might not have the user field
            if 'user' in json_in:
                self.user = json_in['user']
            # if not, make a new token
            else:
                logger.debug(f'Creating new token')
                self.new_token()
        else:
            logger.debug(f'No token exists at {path}. Creating new one.')
            self.new_token()

        # check if the user in the token matches the user in the config
        if self.user != uconfig.get('RunDB', 'rundb_api_user'):
            logger.info(
                f"Username in {uconfig.config_path} does not match token. Overwriting the token.")
            self.new_token()

        # refresh if needed
        if not self.is_valid:
            self.new_token()
            #self.refresh()
        else:
            logger.debug("Token is valid.")

    def __call__(self):
        return self.token_string

    def new_token(self):
        path = PREFIX + "/login"
        username = uconfig.get('RunDB', 'rundb_api_user')
        pw = uconfig.get('RunDB', 'rundb_api_password')
        data = json.dumps({"username": username,
                           "password": pw})
        logger.debug('Creating a new token: doing API call now')
        # try making a new token 3 times
        success = False
        for _try in range(3):
            try:
                response = requests.post(path, data=data, headers=BASE_HEADERS)
                response_json = json.loads(response.text)
                success = True
                break
            except json.decoder.JSONDecodeError:
                logger.info(f"Login attempt #{_try+1} failed. "
                            f"Sleeping for {10**_try} seconds and trying again.")
                time.sleep(10**_try)

        if not success:
            raise NewTokenError("Error in creating a token.")


        logger.debug(f'The response contains these keys: {list(response_json.keys())}')
        token = response_json.get('access_token', 'CALL_FAILED')
        if token == 'CALL_FAILED':
            logging.error(
                f"API call to create new token failed. Here is the response:\n{response.text}")
            raise RuntimeError("Creating a new token failed.")
        self.token_string = token
        self.user = username
        self.creation_time = datetime.datetime.now().timestamp()
        self.write()

    @property
    def is_valid(self):
        # TODO do an API call for this instead?
        diff = datetime.datetime.now().timestamp() - self.creation_time
        return diff < 24 * 60 * 60

    @property
    def json(self):
        return dict(string=self.token_string, creation_time=self.creation_time, user=self.user)

    def refresh(self):
        # update the token string
        url = PREFIX + "/refresh"
        headers = BASE_HEADERS.copy()
        headers['Authorization'] = f"Bearer {self.token_string}"
        logger.debug(f"Refreshing your token with API call {url}")
        response = requests.get(url, headers=headers)
        response_json = json.loads(response.text)
        logger.debug(f'The response contains these keys: {list(response_json.keys())}')
        # if renew fails, try logging back in
        if response.status_code != 200:
            if response_json['error'] != 'EarlyRefreshError':
                logger.warning("Refreshing token failed for some reason, so making a  new one")
                self.new_token()
                self.creation_time = datetime.datetime.now().timestamp()
                logger.debug("Token refreshed")
        else:
            self.creation_time = datetime.datetime.now().timestamp()
        self.write()

    def write(self):
        logger.debug(f"Dumping token to disk at {self.path}.")
        with open(self.path, "w") as f:
            json.dump(self.json, f)


class DB():
    """Wrapper around the RunDB API"""

    def __init__(self, token_path=None):

        if token_path is None:
            if 'HOME' not in os.environ:
                logger.error('$HOME is not defined in the environment')
                if 'USERPROFILE' in os.environ:
                    # Are you on windows?
                    token_path = os.path.join(os.environ['USERPROFILE'], '.dbtoken')
            else:
                token_path = os.path.join(os.environ['HOME'], ".dbtoken")

        # Takes a path to serialized token object
        token = Token(token_path)

        self.headers = BASE_HEADERS.copy()
        self.headers['Authorization'] = "Bearer {token}".format(token=token())

    # Helper:
    @Responder
    def _get(self, url):
        return requests.get(PREFIX + url, headers=self.headers)

    @Responder
    def _put(self, url, data):
        return requests.put(PREFIX + url, data=data, headers=self.headers)

    @Responder
    def _post(self, url, data):
        # DOES THIS WORK?
        return requests.post(PREFIX + url, data=data, headers=self.headers)

    @Responder
    def _delete(self, url, data):
        return requests.delete(PREFIX + url, data=data, headers=self.headers)

    def _is_run_number(self, identifier):
        '''
        Takes a string and classifies it as a run number (as opposed to a
        run name)
        '''
        if re.search('^[0-9]+$', identifier):
            return True
        return False

    def _get_from_results(self, name_or_number, key):
        url = "/runs/number/{name_or_number}/filter/detector".format(name_or_number=name_or_number)
        response = json.loads(self._get(url).text)
        if (response is None
                or 'results' not in response
                or key not in response['results']):
            logger.warning(f'Cannot get {name_or_number} from {url}')
        else:
            return response['results'][key]

    def get_name(self, name):
        return self._get_from_results(name, 'name')

    def get_number(self, number):
        return self._get_from_results(number, 'number')

    def get_did(self, identifier, type='raw_records'):
        doc = self.get_doc(identifier)
        for d in doc['data']:
            if not ('host' in d and 'type' in d and 'did' in d):
                # This ddoc is not in the format of rucio
                continue
            if d['host'] == 'rucio-catalogue' and d['type'] == type:
                return d['did']
        raise ValueError(f'No {identifier} for {type}')

    def get_doc(self, identifier):
        '''
        Retrieves a document from the database. The identifier
        could be a run number of run name - the disambiguation
        takes place automatically.
        '''

        # map from all kinds of types (int, np int, ...)
        identifier = str(identifier)

        url = '/runs/name/{num}'.format(num=identifier)
        if self._is_run_number(identifier):
            url = '/runs/number/{num}'.format(num=identifier)
        # TODO what should be default
        return json.loads(self._get(url).text).get('results', None)

    def get_data(self, identifier):
        '''
        Retrieves the data portion of a document from the
        database. The identifier could be a run number of
        run name - the disambiguation takes place
        automatically.
        '''

        # map from all kinds of types (int, np int, ...)
        identifier = str(identifier)

        url = '/runs/name/{num}/data'.format(num=identifier)
        if self._is_run_number(identifier):
            url = '/runs/number/{num}/data'.format(num=identifier)

        data = json.loads(self._get(url).text).get('results', {})
        if 'data' not in data:
            raise RuntimeError('The requested document does not have a data key/value')

        return data['data']

    def update_data(self, identifier, datum):
        '''
        Updates a data entry. Identifier can be run number of name.
        '''

        # map from all kinds of types (int, np int, ...)
        identifier = str(identifier)

        datum = json.dumps(datum)

        url = '/run/name/{num}/data/'.format(num=identifier)
        if self._is_run_number(identifier):
            url = '/run/number/{num}/data/'.format(num=identifier)

        return self._post(url, data=datum)

    def delete_data(self, identifier, datum):
        '''
        Updates a datum for a document with a matching identifier
        (name or run number)
        '''

        # map from all kinds of types (int, np int, ...)
        identifier = str(identifier)

        datum = json.dumps(datum)

        url = '/run/name/{num}/data/'.format(num=identifier)
        if self._is_run_number(identifier):
            url = '/run/number/{num}/data/'.format(num=identifier)

        return self._delete(url, data=datum)

    def query(self, page_num):
        url = '/runs/page/{page_num}'.format(page_num=page_num)
        response = json.loads(self._get(url).text)
        return response.get('results', {})

    def query_by_source(self, source, page_num):
        url = '/runs/source/{source}/page/{page_num}'.format(source=source, page_num=page_num)
        response = json.loads(self._get(url).text)
        return response.get('results', {})

    def query_by_tag(self, tag, page_num):
        url = '/runs/tag/{tag}/page/{page_num}'.format(tag=tag, page_num=page_num)
        response = json.loads(self._get(url).text)
        return response.get('results', {})

    def get_hash(self, context, datatype, straxen_version):
        if '.' in straxen_version:
            straxen_version = straxen_version.replace('.', '_')
        url = '/contexts/{straxen_version}/{context}/{dtype}'.format(context=context,
                                                                     dtype=datatype,
                                                                     straxen_version=straxen_version)
        response = json.loads(self._get(url).text)
        return response.get('results', {})

    def update_context_collection(self, data):
        context = data.get('name')
        straxen_version = data.get('straxen_version')
        straxen_version = straxen_version.replace('.', '_')
        url = '/contexts/{straxen_version}/{context}/'.format(context=context,
                                                              straxen_version=straxen_version)
        data['date_added'] = data['date_added'].isoformat()
        response = json.loads(self._post(url, data=json.dumps(data)).text)
        return response.get('results', {})

    def delete_context_collection(self, context, straxen_version):
        straxen_version = straxen_version.replace('.', '_')
        url = '/contexts/{straxen_version}/{context}/'.format(context=context,
                                                              straxen_version=straxen_version)
        response = json.loads(self._delete(url, data=None).text)
        return response.get('results', {})

    def get_context(self, context, straxen_version):
        straxen_version = straxen_version.replace('.', '_')
        url = '/contexts/{straxen_version}/{context}/'.format(context=context,
                                                              straxen_version=straxen_version)
        response = json.loads(self._get(url).text)
        return response.get('results', {})

    def get_rses(self, run_number, dtype, hash):
        data = self.get_data(run_number)
        rses = []
        for d in data:
            assert 'host' in d and 'type' in d, (
                f"invalid data-doc retrieved for {run_number} {dtype} {hash}")
            # Did is only in rucio-cataloge, hence don't ask for it to
            # be in all docs in data
            if (d['host'] == "rucio-catalogue" and d['type'] == dtype and
                    hash in d['did'] and d['status'] == 'transferred'):
                rses.append(d['location'])

        return rses

    # TODO
    def get_all_contexts(self):
        """Loads all contexts"""
        raise NotImplementedError

    # TODO
    def get_context_info(self, dtype, strax_hash):
        """Returns context name and strax(en) versions for a given dtype and hash"""
        raise NotImplementedError

    def get_mc_documents(self):
        '''
        Returns all MC documents.
        '''
        url = '/mc/documents/'
        return self._get(url)

    def add_mc_document(self, document):
        '''
        Adds a document to the MC database.
        '''
        doc = json.dumps(document)
        url = '/mc/documents/'
        return self._post(url, data=doc)

    def delete_mc_document(self, document):
        '''
        Deletes a document from the MC database. The document must be passed exactly.
        '''
        doc = json.dumps(document)
        url = '/mc/documents/'
        return self._delete(url, data=doc)

    def download_file(self, filename, save_dir='./', force=False):
        """Downloads file from GridFS"""
        url = f'/files/{filename}'
        os.makedirs(save_dir, exist_ok=True)
        write_to = os.path.join(save_dir, filename)
        if os.path.exists(write_to) and not force:
            logger.debug(f"{filename} already exists at {write_to} and the 'force' flag is not set.")
        else:
            logger.debug(f"Downloading {filename} from gridfs...")
            response = self._get(url)
            with open(write_to, 'wb') as f:
                f.write(response.content)
            logger.debug(f'DONE. {filename} downloaded to {write_to}')
        return write_to

    def load_file(self, filename, save_dir=None, force=False):
        if save_dir is None:
            save_dir = os.path.join(os.environ.get("HOME"), '.gridfs_cache')
        path = self.download_file(filename, save_dir=save_dir, force=force)
        return io.read_file(path)

    def upload_file(self, filepath, filename=None):
        with open(filepath, 'rb') as f:
            fb = f.read()
        # if no specific filename passed, just get it from the path
        if not filename:
            filename = os.path.basename(filepath)
        url = f'/files/{filename}'
        return self._post(url, data=fb)

    def get_files(self, query: dict, projection=None):
        """Do a general query on the fs.files collection"""
        url = '/files/query'
        # the projection needs to be a dict for the flask app
        if projection is None:
            projection = {}
        data = json.dumps(dict(query=query, projection=projection))
        result = self._post(url, data=data)
        response = result.json()
        return response.get('results', [])

    def count_files(self, query: dict)->int:
        """Perform colection.count_documents on the fs.files-collection using the query"""
        """<URL MAGIC>"""
        docs = self.get_files(query)
        return len(docs)

    def delete_file(self, filename):
        resp = input(f"HUGE GIGANTIC CRITICAL WARNING: this will delete all files of the name {filename} in GridFS. "
                     "Confirm by typing again the name of the file you want to delete: \n")
        if resp != filename:
            print("Probably the safe choice. Exiting.")
            return
        url = f'/files/{filename}'
        return self._delete(url, data="")

    def get_file_md5(self, filename):
        url = f"/files/{filename}/md5"
        response = self._get(url).json()
        return response['results']



class PyMongoCannotConnect(Exception):
    """Raise error when we cannot connect to the pymongo client"""
    pass


def test_collection(collection, url, raise_errors=False):
    """
    Warn user if client can be troublesome if read preference is not specified
    :param collection: pymongo client
    :param url: the mongo url we are testing (for the error message)
    :param raise_errors: if False (default) warn, otherwise raise an error
    """
    try:
        # test the collection by doing a light query
        collection.find_one({}, {'_id': 1})
    except (pymongo.errors.ServerSelectionTimeoutError, pymongo.errors.OperationFailure) as e:
        # This happens when trying to connect to one or more mirrors
        # where we cannot decide on who is primary
        message = (
            f'Cannot get server info from "{url}". Check your config at {uconfig.config_path}')
        if not raise_errors:
            warn(message)
        else:
            message += (
                'This usually happens when trying to connect to multiple '
                'mirrors when they cannot decide which is primary. Also see:\n'
                'https://github.com/XENONnT/straxen/pull/163#issuecomment-732031099')
            raise PyMongoCannotConnect(message) from e


def pymongo_collection(collection='runs', **kwargs):
    # default collection is the XENONnT runsDB
    # for 1T, pass collection='runs_new'
    print("WARNING: pymongo_collection is deprecated. Please use xent_collection or xe1t_collection instead")
    uri = 'mongodb://{user}:{pw}@{url}'
    url = kwargs.get('url')
    user = kwargs.get('user')
    pw = kwargs.get('password')
    database = kwargs.get('database')

    if not url:
        url = uconfig.get('RunDB', 'pymongo_url')
    if not user:
        user = uconfig.get('RunDB', 'pymongo_user')
    if not pw:
        pw = uconfig.get('RunDB', 'pymongo_password')
    if not database:
        database = uconfig.get('RunDB', 'pymongo_database')
    uri = uri.format(user=user, pw=pw, url=url)
    c = pymongo.MongoClient(uri, readPreference='secondaryPreferred')
    DB = c[database]
    coll = DB[collection]
    # Checkout the collection we are returning and raise errors if you want
    # to be realy sure we can use this URL.
    # test_collection(coll, url, raise_errors=False)

    return coll


def _collection(experiment, collection, url=None, user=None, password=None, database=None):
    if experiment not in ['xe1t', 'xent']:
        raise ValueError(f"experiment must be 'xe1t' or 'xent'. You passed f{experiment}")

    if not url:
        url = uconfig.get('RunDB', f'{experiment}_url')
    if not user:
        user = uconfig.get('RunDB', f'{experiment}_user')
    if not password:
        password = uconfig.get('RunDB', f'{experiment}_password')
    if not database:
        database = uconfig.get('RunDB', f'{experiment}_database')

    uri = f"mongodb://{user}:{password}@{url}"
    c = pymongo.MongoClient(uri, readPreference='secondaryPreferred')
    DB = c[database]
    coll = DB[collection]
    return coll


def xent_collection(collection='runs', **kwargs):
    return _collection('xent', collection, **kwargs)


def xe1t_collection(collection='runs_new', **kwargs):
    return _collection('xe1t', collection, **kwargs)



import numpy as np
import pickle
import gzip
import json
import pandas as pd
import os
import commentjson


def read_file(path):
    """
    Open a file from disk. Auto-infers the file format
    :param path: str, file to open
    :return: opened file
    """
    # copied from straxen.common.open_resource
    # https://github.com/XENONnT/straxen/blob/a2e0e3abdbf278000cda70f7662a7d841c7223ef/straxen/common.py#L85
    name, fmt = os.path.splitext(path)

    if fmt in ['.npy', '.npy_pickle', '.npz']:
        result = np.load(path, allow_pickle=fmt == 'npy_pickle')
        if isinstance(result, np.lib.npyio.NpzFile):
            # Slurp the arrays in the file, so the result can be copied,
            # then close the file so its descriptors does not leak.
            result_slurped = {k: v[:] for k, v in result.items()}
            result.close()
            result = result_slurped
    elif fmt == '.pkl':
        with open(path, 'rb') as f:
            result = pickle.load(f)
    elif fmt == '.gz':
        subname, subfmt = os.path.splitext(name)
        if subfmt == '.pkl':
            with gzip.open(path, 'rb') as f:
                result = pickle.load(f)
        elif subfmt == '.json':
            with gzip.open(path, 'rb') as f:
                result = json.load(f)
    elif fmt == '.json':
        with open(path, mode='r') as f:
            result = commentjson.load(f)
    elif fmt == '.binary':
        with open(path, mode='rb') as f:
            result = f.read()
    elif fmt == '.text':
        with open(path, mode='r') as f:
            result = f.read()
    elif fmt == '.csv':
        result = pd.read_csv(path)
    else:
        raise ValueError(f"Unsupported format {fmt}!")

    return result
